/* Copyright (c) 2003-2012 Synology Inc. All rights reserved. */
#include <synosdk/file.h>
#include <synodns.h>
#include <LocalizedStringParser.h>
#include <getopt.h>
#include <synosdk/copyright.h>
#include <synosdk/conf.h>
#include <synosdk/proc.h>

#include <unistd.h>
#include <sys/file.h>

using namespace std;

#define SZF_SYNODNSTOOL_LOCK_FILE "/tmp/synodnstool.lock"

class Locker {
private:
	int _fdLock;
	bool _blLocked;
public:
	Locker(const char* szLockFileName) {
		_fdLock = -1;
		if (szLockFileName && ('\0' != szLockFileName[0])) {
			_fdLock = open(szLockFileName, O_RDWR | O_CREAT, 0644); // blocking lock
			_blLocked = true;
			if ((0 > _fdLock) || (0 > flock(_fdLock, LOCK_EX))) {
				_blLocked = false;
				SYSLOG(LOG_ERR, "lock %s failed", szLockFileName);
			}
		}
	}
	bool isLocked() {
		return ((0 <= _fdLock) && _blLocked);
	}
	~Locker() {
		if (0 <= _fdLock) {
			flock(_fdLock, LOCK_UN);
			close(_fdLock);
			_fdLock = -1;
		}
	}
};

typedef enum __TAG_SYNO_DNS_TOOL_ACTION {
	NONE,
	SYNO_DNS_TOOL_ACTION_SET_DNSMASQ_ALLOW_RECURSION_LIST,
	SYNO_DNS_TOOL_ACTION_CLEAR_DNSMASQ_ALLOW_RECURSION_LIST,
	SYNO_DNS_TOOL_ACTION_APPLY_LOG_CONF,
	SYNO_DNS_TOOL_ACTION_FIX_REDEFINED,
	SYNO_DNS_TOOL_ACTION_ENABLE_RECURSION,
	SYNO_DNS_TOOL_ACTION_LOG_DB_INIT,
	SYNO_DNS_TOOL_ACTION_HELP
} SYNO_DNS_TOOL_ACTION;

#define SZ_HELP "help"
#define SZ_SET_DNS_MASQ_ALLOW_RECURSION "set-dnsmasq-allow-recursion"
#define SZ_CLEAR_DNS_MASQ_ALLOW_RECURSION "clear-dnsmasq-allow-recursion"
#define SZ_APPLY_LOG_CONF "apply-log-conf"
#define SZ_FIX_REDEFINED "fix-redefined" //B#219
#define SZ_ENABLE_RECUSION "enable-recursion" //B#650
#define SZ_LOG_DB_INIT "log-db-init"

static struct option longopts[] = {
	{SZ_HELP,								no_argument,        0, 	SYNO_DNS_TOOL_ACTION_HELP},
	{SZ_SET_DNS_MASQ_ALLOW_RECURSION, 	     required_argument, 0,  SYNO_DNS_TOOL_ACTION_SET_DNSMASQ_ALLOW_RECURSION_LIST},
	{SZ_CLEAR_DNS_MASQ_ALLOW_RECURSION,    	 no_argument,		0,  SYNO_DNS_TOOL_ACTION_CLEAR_DNSMASQ_ALLOW_RECURSION_LIST},
	{SZ_APPLY_LOG_CONF, no_argument, 0, SYNO_DNS_TOOL_ACTION_APPLY_LOG_CONF},
	{SZ_FIX_REDEFINED, no_argument, 0, SYNO_DNS_TOOL_ACTION_FIX_REDEFINED},
	{SZ_ENABLE_RECUSION, no_argument, 0, SYNO_DNS_TOOL_ACTION_ENABLE_RECURSION},
	{SZ_LOG_DB_INIT, no_argument, 0, SYNO_DNS_TOOL_ACTION_LOG_DB_INIT},
	{NULL, 0, NULL, 0}
};

void usage() 
{
	int nOptions = sizeof(longopts)/sizeof(struct option) -1;
	SLIBCopyright();
	printf("Usage: synodnstool\n");
	for (int i = 0; i < nOptions; i++) {
		printf ("\t--");
		printf ("%s", longopts[i].name);
		if (0 == strcmp(SZ_SET_DNS_MASQ_ALLOW_RECURSION, longopts[i].name)) {
			printf (" [allow-recusrsion ip address]");
		}
		printf ("\n");
	}
}

static int SetDNSMasqList(const string &strIP) 
{
	int nRet = -1;
	PSYNODNSRESOLVECONF pDnsResolveConf = NULL;

	SLIBCFileSetSectionValue(SZF_DNS_CONF, SZ_DNS_RESOLVE_SERVICE, SZK_DNS_MASQ_RECURSION, strIP.c_str(), "\t%s=\"%s\"\n");

	if (!(pDnsResolveConf = SYNODnsResolveConfGet())) {
		SYSLOG(LOG_ERR, "Fail to get dns configuration. szFile=[%s],\
		synoerr=[0x%04X]", SZF_DNS_CONF, SLIBCErrGet());
		goto End;
	}

	if (0 > SYNODnsResolveConfSet(pDnsResolveConf)) {
		SYSLOG(LOG_ERR, "SYNODnsResolveConfSet failed");
		goto End;
	}

	nRet = 0;
End:
	SYNODnsResolveConfFree(pDnsResolveConf);
	return nRet;
}

static int ClearDNSMasqList()
{
	int nRet = -1;
	PSYNODNSRESOLVECONF pDnsResolveConf = NULL;

	if (0 > SLIBCFileRemoveKey(SZF_DNS_CONF, SZK_DNS_MASQ_RECURSION)) {
		SYSLOG(LOG_ERR, "SLIBCFileRemoveKey(%s, %s) failed", SZF_DNS_CONF, SZK_DNS_MASQ_RECURSION);
		goto End;
	}

	if (!(pDnsResolveConf = SYNODnsResolveConfGet())) {
		SYSLOG(LOG_ERR, "Fail to get dns configuration. szFile=[%s],\
		synoerr=[0x%04X]", SZF_DNS_CONF, SLIBCErrGet());
		goto End;
	}

	if (0 > SYNODnsResolveConfSet(pDnsResolveConf)) {
		SYSLOG(LOG_ERR, "SYNODnsResolveConfSet failed");
		goto End;
	}

	nRet = 0;
End:
	SYNODnsResolveConfFree(pDnsResolveConf);
	return nRet;
}

static int FixRedefined() {
	int nRet = -1;

	int nAllowRecursion = 0;
	int nForwarders = 0;
	int nForward = 0;
	size_t nLen = 0;
	ssize_t nRead = 0;
	char *szTmp = NULL;

	char szDNSMasqList[BUFSIZ] = {0}; // can save about 200 subnet or ip

//1. check if any key work redefined in SZF_DNS_RSV_CONF ?
//2. if yes, need to fix it by rewrite SZF_DNS_RSV_CONF
	FILE *pFile = NULL;
	pFile = fopen(SZF_DNS_RSV_CONF, "r");
	if (NULL == pFile) {
		SYSLOG(LOG_ERR, "pFile is NULL");
		goto End;
	}

	while (-1 != (nRead = getline(&szTmp, &nLen, pFile))) {
		if (NULL != strstr(szTmp, "allow-recursion ")) {
			nAllowRecursion++;
		} else if (NULL != strstr(szTmp, "forwarders ")) {
			nForwarders++;
		} else if (NULL != strstr(szTmp, "forward ")) {
			nForward++;
		}

		if ((2 <= nAllowRecursion) || 
				(2 <= nForwarders) ||
				(2 <= nForward)) {
			SYSLOG(LOG_ERR, "%s redefined occurs... try to fix it", SZF_DNS_RSV_CONF);
	
			if (0 == SLIBCFileGetSectionValue(SZF_DNS_CONF, SZ_DNS_RESOLVE_SERVICE, SZK_DNS_MASQ_RECURSION, szDNSMasqList, sizeof(szDNSMasqList))) {
				if (0 > SetDNSMasqList(szDNSMasqList)) {
					SYSLOG(LOG_ERR, "SetDNSMasqList(%s) failed", szDNSMasqList);
				}
			} else { // get failed, still need to clear
				if (0 > ClearDNSMasqList()) {
					SYSLOG(LOG_ERR, "ClearDNSMasqList() failed");
				}
			}
			break;
		}
	}

	fclose(pFile);

	nRet = 0;
End:
	if (szTmp) {
		free(szTmp);
	}
	return nRet;
}

static int EnableRecursion()
{
	int ret = -1;
	PSYNODNSRESOLVECONF pDnsResolveConf = NULL;

	if (!(pDnsResolveConf = SYNODnsResolveConfGet())) {
		SYSLOG(LOG_ERR, "SYNODnsResolveConfGet failed, synoerr=[0x%04X]", SLIBCErrGet());
		goto End;
	}

	if (!pDnsResolveConf->resolve_enable) {
		pDnsResolveConf->resolve_enable = true;
		pDnsResolveConf->limit_enable = true;
	}

	if (0 > SYNODnsResolveConfSet(pDnsResolveConf)) {
		SYSLOG(LOG_ERR, "SYNODnsResolveConfSet failed, synoerr=[0x%04X] ", SLIBCErrGet());
		goto End;
	}

	ret = 0;
End:
	SYNODnsResolveConfFree(pDnsResolveConf);
	return ret;
}

static int LogDBInit(void)
{
	int nRet = -1;
	char szTmp[512] = {'\0'};
	int nTmp = 0;

	if (0 > SLIBCFileGetSectionValue(SZF_DNS_CONF, SZ_DNS_ADVANCED_SERVICE, SZK_DNS_LOG_SPACE, szTmp, sizeof(szTmp))) {
		SYSLOG(LOG_ERR, "Fail to get dns configuration!! szFile=[%s], synoerr=[0x%04X]", SZF_DNS_CONF, SLIBCErrGet());
		goto End;
	}
	if (0 > (nTmp = SYNODnsToInt(szTmp))) {
		nTmp = 0;
	}
	if (0 > SYNODNSLogDBSizeSet(nTmp)) {
		SYSLOG(LOG_ERR, "SYNODNSLogDBInit failed");
		goto End;
	}
	if (0 != SLIBCExec("/bin/chown", "nobody:root", SZF_DNS_LOG_DB, NULL, NULL)) {
		SYSLOG(LOG_ERR, "chown [%s] failed", SZF_DNS_LOG_DB);
		goto End;
	}
	if (0 != SLIBCExec("/bin/chmod", "0664", SZF_DNS_LOG_DB, NULL, NULL)) {
		SYSLOG(LOG_ERR, "chmod [%s] failed", SZF_DNS_LOG_DB);
		goto End;
	}
	nRet = 0;
End:
	return nRet;
}

int main(int argc, char* argv[]) {
	int nRet = 1;
	int long_index = 0;
	char szIP[BUFSIZ] = {};
	SYNO_DNS_TOOL_ACTION action = NONE;
	SYNO_DNS_TOOL_ACTION tmpAction = NONE;
	//to avoid shell script hook enter this function to write SZF_DNS_CONF SZ_DNS_RESOLVE_SERVICE concurentlly
	Locker lock(SZF_SYNODNSTOOL_LOCK_FILE);
	if (!lock.isLocked()) {
		SYSLOG(LOG_ERR, "lock %s failed", SZF_SYNODNSTOOL_LOCK_FILE);
		goto End;
	}

	if ((tmpAction = (SYNO_DNS_TOOL_ACTION)getopt_long_only(argc, argv, "", longopts, &long_index)) != -1) {
		switch (tmpAction) {		
			case SYNO_DNS_TOOL_ACTION_SET_DNSMASQ_ALLOW_RECURSION_LIST:
				snprintf(szIP, sizeof(szIP), "%s", optarg);
				action = tmpAction;
				break;
			case SYNO_DNS_TOOL_ACTION_CLEAR_DNSMASQ_ALLOW_RECURSION_LIST:
			case SYNO_DNS_TOOL_ACTION_APPLY_LOG_CONF:
			case SYNO_DNS_TOOL_ACTION_FIX_REDEFINED:
			case SYNO_DNS_TOOL_ACTION_ENABLE_RECURSION:
			case SYNO_DNS_TOOL_ACTION_LOG_DB_INIT:
				action = tmpAction;
				break;
			case SYNO_DNS_TOOL_ACTION_HELP:
			default:
				usage();
				goto End;
		}
	} 

	//actions
	switch (action) {
		case SYNO_DNS_TOOL_ACTION_SET_DNSMASQ_ALLOW_RECURSION_LIST:
			if (0 > SetDNSMasqList(szIP)) {
				SYSLOG(LOG_ERR, "SetDNSMasqList(%s) failed", szIP);
				goto End;
			}
			break;
		case SYNO_DNS_TOOL_ACTION_CLEAR_DNSMASQ_ALLOW_RECURSION_LIST:
			if (0 > ClearDNSMasqList()) {
				SYSLOG(LOG_ERR, "ClearDNSMasqList() failed");
				goto End;
			}
			break;
		case SYNO_DNS_TOOL_ACTION_APPLY_LOG_CONF:
			if (0 > SYNODnsLogConfApply()) {
				SYSLOG(LOG_ERR, "SYNODnsLogConfApply failed");
				goto End;
			}
			break;
		case SYNO_DNS_TOOL_ACTION_FIX_REDEFINED:
			if (0 > FixRedefined()) {
				SYSLOG(LOG_ERR, "FixRedefined() failed");
				goto End;
			}
			break;
		case SYNO_DNS_TOOL_ACTION_ENABLE_RECURSION:
			if(0 > EnableRecursion()) {
				SYSLOG(LOG_ERR, "EnableRecursion failed");
				goto End;
			}
			if (SLIBCBlProcessIsAlive("named") && (0 > SYNODNSServerReload())) {
				SYSLOG(LOG_ERR, "SYNODNSServerReload failed");
			}
			break;
		case SYNO_DNS_TOOL_ACTION_LOG_DB_INIT:
			if (0 > LogDBInit()) {
				goto End;
			}
			break;
		default:
			printf ("Error: no such action\n");
			usage();
			goto End;
	}

	nRet = 0;
End:
	return nRet;
}
